from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
import threading
from typing import Callable, Final, Sequence, TYPE_CHECKING
import weakref

from .connection import Connection, InvalidStateError, MessageHandlers
from .error import MQTTError
from .handles import SubscribeHandle, UnsubscribeHandle
from .logger import get_logger
from .mqtt_spec import MAX_PACKET_ID, MQTTReasonCode, MQTTQoS
from .packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from .property import MQTTSubscribeProps, MQTTUnsubscribeProps
from .topic_alias import InboundTopicAlias
from .topic_filter import validate_topic_filter, validate_share_name, join_share, match_topic_filter

if TYPE_CHECKING:
    from .client import Client

logger: Final = get_logger("subscriptions")

SubscribeCallback = Callable[["Client", MQTTPublishPacket], None]


class NoMatchingSubscriptionError(Exception):
    """Exception raised when no matching subscription is found when unsubscribing."""


class ReconfiguredError(Exception):
    """A subscribe operation was reconfigured before being acknowledged."""


class RetainPolicy(IntEnum):
    """Policy for broker sending retained messages upon a subscription.

    ALWAYS: Always send retained messages on subscription.

    ONCE: Only send retained messages on the first subscription to a topic.

    NEVER: Never send retained messages on subscription."""
    ALWAYS = 0
    ONCE = 1
    NEVER = 2


class _SubscriptionState(IntEnum):
    """Internal state of a subscription."""
    SUBSCRIBING = 0
    SUBSCRIBED = 1
    UNSUBSCRIBING = 2
    UNSUBSCRIBED = 3


@dataclass(match_args=True, slots=True)
class Subscription:
    """All the data about a request for subscription."""
    topic_filter: str
    callback: SubscribeCallback
    max_qos: MQTTQoS = MQTTQoS.Q2
    share_name: str | None = None
    no_local: bool = False
    retain_as_published: bool = False
    retain_policy: RetainPolicy = RetainPolicy.ALWAYS
    sub_id: int | None = None
    user_properties: tuple[tuple[str, str], ...] | None = None
    state: _SubscriptionState = field(default=_SubscriptionState.SUBSCRIBING)
    effective_filter: str = field(init=False, repr=False)

    def __post_init__(self) -> None:
        validate_topic_filter(self.topic_filter)
        if self.share_name is not None:
            validate_share_name(self.share_name)
        self.effective_filter = join_share(self.topic_filter, self.share_name)

    def render_sub(self) -> MQTTSubscribePacket:
        """Render the subscription into a SUBSCRIBE packet."""
        opts = self.max_qos.value | (self.retain_policy << 4) | (self.retain_as_published << 3) | (self.no_local << 2)
        props = MQTTSubscribeProps()
        if self.sub_id is not None:
            props.SubscriptionIdentifier = {self.sub_id}
        if self.user_properties is not None:
            props.UserProperty = self.user_properties
        return MQTTSubscribePacket(
            topics=[(self.effective_filter, opts)],
            properties=props,
        )

    def render_unsub(self) -> MQTTUnsubscribePacket:
        """Render the subscription into an UNSUBSCRIBE packet."""
        props = MQTTUnsubscribeProps()
        return MQTTUnsubscribePacket(
            topics=[self.effective_filter],
            properties=props,
        )


class Subscriptions:
    """Container for MQTT subscriptions and their callbacks."""
    __slots__ = (
        "__weakref__",
        "_client",
        "_cond",
        "_connection",
        "_inflight_sub_packet_ids",
        "_inflight_unsub_packet_ids",
        "_next_sub_packet_id",
        "_next_unsub_packet_id",
        "_sub_handles",
        "_subscriptions",
        "_topic_alias",
        "_unsub_handles",
    )

    def __init__(
        self,
        handlers: MessageHandlers,
        connection: Connection,
        client: weakref.ReferenceType[Client],
    ) -> None:
        self._connection = connection
        self._client = client
        self._cond = threading.Condition(connection.fsm.lock)
        self._topic_alias = InboundTopicAlias()
        self._subscriptions: dict[str, Subscription] = {}
        self._sub_handles: weakref.WeakValueDictionary[str, SubscribeHandle] = weakref.WeakValueDictionary()
        self._unsub_handles: weakref.WeakValueDictionary[str, UnsubscribeHandle] = weakref.WeakValueDictionary()
        self._inflight_sub_packet_ids: dict[int, str] = {}
        self._inflight_unsub_packet_ids: dict[int, str] = {}
        self._next_sub_packet_id = 1
        self._next_unsub_packet_id = 1

        handlers.register(MQTTSubAckPacket, self.handle_suback)
        handlers.register(MQTTUnsubAckPacket, self.handle_unsuback)

    def subscribe(
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: MQTTQoS = MQTTQoS.Q2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
    ) -> SubscribeHandle:
        """Add a subscription with a callback.

        Repeated calls to subscribe with the same combination of share_name and topic_filter will reconfigure
        the existing subscription with the new parameters, discarding the previous callback.

        :return: A SubscribeHandle which can be used to wait for ack."""
        sub = Subscription(
            topic_filter=topic_filter,
            callback=callback,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=tuple(user_properties) if user_properties is not None else None,
        )
        with self._cond:
            self._subscriptions[sub.effective_filter] = sub
            if (existing := self._sub_handles.get(sub.effective_filter, None)) is not None and existing.ack is None and existing.exc is None:
                existing.exc = ReconfiguredError("Topic was reconfigured")
                self._cond.notify_all()
            handle = SubscribeHandle(self._cond)
            self._sub_handles[sub.effective_filter] = handle
            self._flush_packets()
            return handle

    def _flush_packets(self) -> None:
        """Try to send all pending SUBSCRIBE and UNSUBSCRIBE packets."""
        with self._cond:
            pending_subs = [
                sub for sub in self._subscriptions.values() if (
                    sub.state == _SubscriptionState.SUBSCRIBING and sub.effective_filter not in self._inflight_sub_packet_ids.values()
                )
            ]
            for sub in pending_subs:
                sub_packet = sub.render_sub()
                sub_packet.packet_id = self._get_next_sub_packet_id()
                self._inflight_sub_packet_ids[sub_packet.packet_id] = sub.effective_filter
                try:
                    self._connection.send(sub_packet)
                except InvalidStateError:
                    logger.debug("Connection closed, SUBSCRIBE not sent")
                    del self._inflight_sub_packet_ids[sub_packet.packet_id]
                    return

            pending_unsubs = [
                sub for sub in self._subscriptions.values() if (
                    sub.state == _SubscriptionState.UNSUBSCRIBING and sub.effective_filter not in self._inflight_unsub_packet_ids.values()
                )
            ]
            for sub in pending_unsubs:
                unsub_packet = sub.render_unsub()
                unsub_packet.packet_id = self._get_next_unsub_packet_id()
                self._inflight_unsub_packet_ids[unsub_packet.packet_id] = sub.effective_filter
                try:
                    self._connection.send(unsub_packet)
                except InvalidStateError:
                    logger.debug("Connection closed, UNSUBSCRIBE not sent")
                    del self._inflight_unsub_packet_ids[unsub_packet.packet_id]
                    return

    def _get_next_sub_packet_id(self) -> int:
        """Get the next SUBSCRIBE packet identifier."""
        sub_id = self._next_sub_packet_id
        self._next_sub_packet_id += 1
        if self._next_sub_packet_id > MAX_PACKET_ID:
            self._next_sub_packet_id = 1
        return sub_id

    def unsubscribe(
        self,
        topic_filter: str,
        *,
        share_name: str | None = None,
    ) -> UnsubscribeHandle | None:
        """Unsubscribe from a topic filter.

        :return: UnsubscribeHandle if the topic was subscribed, otherwise None."""
        effective_filter = join_share(topic_filter, share_name)
        with self._cond:
            if effective_filter not in self._subscriptions:
                # Nothing to unsubscribe.
                return None
            sub = self._subscriptions[effective_filter]
            if sub.state == _SubscriptionState.UNSUBSCRIBED:
                return None
            sub.state = _SubscriptionState.UNSUBSCRIBING
            if (existing := self._unsub_handles.get(effective_filter, None)) is not None and existing.ack is None and existing.exc is None:
                existing.exc = ReconfiguredError("Topic was reconfigured")
                self._cond.notify_all()
            handle = UnsubscribeHandle(self._cond)
            self._unsub_handles[effective_filter] = handle
            self._flush_packets()
            return handle

    def _get_next_unsub_packet_id(self) -> int:
        """Get the next UNSUBSCRIBE packet identifier."""
        unsub_id = self._next_unsub_packet_id
        self._next_unsub_packet_id += 1
        if self._next_unsub_packet_id > MAX_PACKET_ID:
            self._next_unsub_packet_id = 1
        return unsub_id

    def handle_publish(self, packet: MQTTPublishPacket) -> None:
        """Handle incoming PUBLISH packets."""
        with self._cond:
            client = self._client()
            if client is None:
                raise RuntimeError("Client went out of scope")
            self._topic_alias.handle(packet)
            subs = [sub for sub in self._subscriptions.values() if match_topic_filter(sub.topic_filter, packet.topic)]
            if packet.properties.SubscriptionIdentifier is not None:
                subs = [sub for sub in subs if sub.sub_id in packet.properties.SubscriptionIdentifier]
            for sub in subs:
                try:
                    sub.callback(client, packet)
                except Exception:  # noqa: PERF203
                    logger.exception("Unhandled exception in subscription callback")

    def handle_suback(self, packet: MQTTSubAckPacket) -> None:
        """Handle incoming SUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in SUBACK return: %s", errs)
        with self._cond:
            effective_topic_filter = self._inflight_sub_packet_ids.pop(packet.packet_id, None)
            if effective_topic_filter is None:
                raise MQTTError(f"Received SUBACK for unknown packet ID: {packet.packet_id}", MQTTReasonCode.ProtocolError)
            sub = self._subscriptions.get(effective_topic_filter, None)
            if sub is not None and sub.state == _SubscriptionState.SUBSCRIBING:
                sub.state = _SubscriptionState.SUBSCRIBED
            handle = self._sub_handles.pop(effective_topic_filter, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_unsuback(self, packet: MQTTUnsubAckPacket) -> None:
        """Handle incoming UNSUBACK packets."""
        if any(True for code in packet.reason_codes if code.is_error()):
            errs = [hex(code) for code in packet.reason_codes if code.is_error()]
            logger.error("Errors found in UNSUBACK return: %s", errs)
        with self._cond:
            effective_topic_filter = self._inflight_unsub_packet_ids.pop(packet.packet_id, None)
            if effective_topic_filter is None:
                raise MQTTError(f"Received UNSUBACK for unknown packet ID: {packet.packet_id}", MQTTReasonCode.ProtocolError)
            sub = self._subscriptions.get(effective_topic_filter, None)
            if sub is not None and sub.state == _SubscriptionState.UNSUBSCRIBING:
                sub.state = _SubscriptionState.UNSUBSCRIBED
            handle = self._unsub_handles.pop(effective_topic_filter, None)
            if handle is not None:
                handle.ack = packet
                self._cond.notify_all()

    def handle_connack(self, packet: MQTTConnAckPacket) -> None:
        """Handle incoming CONNACK packets."""
        with self._cond:
            self._next_sub_packet_id = 1
            self._next_unsub_packet_id = 1
            self._topic_alias.reset()
            self._inflight_sub_packet_ids.clear()
            self._inflight_unsub_packet_ids.clear()
            if not packet.session_present:
                for sub in self._subscriptions.values():
                    if sub.state == _SubscriptionState.SUBSCRIBED:
                        sub.state = _SubscriptionState.SUBSCRIBING
                    elif sub.state == _SubscriptionState.UNSUBSCRIBING:
                        sub.state = _SubscriptionState.UNSUBSCRIBED
            self._flush_packets()



from __future__ import annotations

from typing import Final, TYPE_CHECKING

from .base import FSMState
from ..types import ConnectParams, StateData, StateEnvironment
from ...logger import get_logger

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.shutdown")


class ShutdownState(FSMState):
    """The final state.

    All buffers will be flushed and the socket will be closed immediately on entry."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        # We can enter this state from any other state.
        # Free up as many resources as possible.
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        state_data.ws_decoder.reset()
        state_data.ws_handshake_buffer.clear()
        state_data.ws_nonce = ""
        state_data.write_buffer.clear()
        with fsm.selector:
            env.packet_buffer.clear()
            fsm.selector.close()



import argparse
from functools import partial
import logging
import subprocess
import threading
import time
from typing import TypeAlias

import pytest
from pytest_mock import MockerFixture

from ohmqtt import __version__, MQTTPublishPacket
from ohmqtt.cli import main
from ohmqtt.cli.publish import PublishCommand
from ohmqtt.cli.subscribe import SubscribeCommand
from ohmqtt.cli.common import get_client, BrokerConnectionError
from .util.fake_broker import FakeBroker


CapSysT: TypeAlias = pytest.CaptureFixture[str]


def test_cli_main_entrypoint() -> None:
    result = subprocess.run(
        ["python", "-m", "ohmqtt", "--version"],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    assert __version__ in result.stdout


def test_cli_main_no_args(capsys: CapSysT) -> None:
    main([])
    captured = capsys.readouterr()
    assert "ohmqtt Command Line Interface" in captured.out
    assert "Available commands" in captured.out


def test_cli_main_version(capsys: CapSysT) -> None:
    with pytest.raises(SystemExit):
        main(["--version"])
    captured = capsys.readouterr()
    assert __version__ in captured.out


@pytest.mark.parametrize("arg", ["-v", "--verbose"])
def test_cli_main_verbose(mocker: MockerFixture, arg: str) -> None:
    mock_logging_basic_config = mocker.patch("logging.basicConfig")
    main([arg])
    mock_logging_basic_config.assert_called_once_with(level=logging.DEBUG)


def test_cli_main_publish(mocker: MockerFixture) -> None:
    mock_publish_execute = mocker.patch(
        "ohmqtt.cli.publish.PublishCommand.execute"
    )
    main(["publish", "test/topic", "Hello, MQTT!"])
    mock_publish_execute.assert_called_once()
    args = mock_publish_execute.call_args[0][0]
    assert args.topic == "test/topic"
    assert args.message == "Hello, MQTT!"


@pytest.mark.parametrize("password", [None, "secret"])
def test_cli_command_get_client(mocker: MockerFixture, password: str | None) -> None:
    args = argparse.Namespace()
    args.clean_start = False
    args.client_id = ""
    args.connect_timeout = 1
    args.topic_alias_maximum = 0xffff  # Set any property to test the code path
    args.username = "foo"
    args.password = password

    with FakeBroker() as broker:
        args.address = f"localhost:{broker.port}"
        client = get_client(args)
        assert client.is_connected()
        client.shutdown()

    mocker.patch(
        "ohmqtt.cli.common.Client.loop_until_connected", return_value=False
    )
    with FakeBroker() as broker:
        args.address = f"localhost:{broker.port}"
        with pytest.raises(BrokerConnectionError):
            client = get_client(args)

    args.address = "invalid_address"
    with pytest.raises(BrokerConnectionError):
        get_client(args)


@pytest.mark.parametrize("correlation_data", [None, "asdf"])
@pytest.mark.parametrize("qos", [0, 1, 2])
def test_cli_command_publish(capsys: CapSysT, mocker: MockerFixture, correlation_data: str | None, qos: int) -> None:
    args = argparse.Namespace()
    args.address = "localhost"
    args.topic = "test/topic"
    args.message = "Hello, MQTT!"
    args.qos = qos
    args.retain = False

    if correlation_data is not None:
        args.correlation_data = correlation_data

    mock_client = mocker.MagicMock()
    mock_get_client = mocker.patch(
        "ohmqtt.cli.publish.get_client", return_value=mock_client
    )

    PublishCommand.execute(args)

    mock_get_client.assert_called_once_with(args)
    mock_client.publish.assert_called_once_with(
        "test/topic",
        b"Hello, MQTT!",
        qos=qos,
        retain=False,
        properties=mocker.ANY,
    )
    if correlation_data is not None:
        called_props = mock_client.publish.call_args[1]["properties"]
        assert called_props.CorrelationData == correlation_data.encode("utf-8")

    captured = capsys.readouterr()
    assert "Published message to test/topic" in captured.out


def test_cli_command_subscribe(capsys: CapSysT, mocker: MockerFixture) -> None:
    class MockArgs(argparse.Namespace):
        address = "localhost"
        topic = "test/topic"
        max_qos = 2
        no_retained = False

    mock_client = mocker.MagicMock()
    mock_get_client = mocker.patch(
        "ohmqtt.cli.subscribe.get_client", return_value=mock_client
    )

    # Mock the loop_forever method to raise KeyboardInterrupt soon
    def mock_loop_forever() -> None:
        time.sleep(1.0)
        raise KeyboardInterrupt

    mock_client.loop_forever.side_effect = mock_loop_forever

    thread = threading.Thread(
        target=partial(SubscribeCommand.execute, MockArgs())
    )
    thread.start()

    time.sleep(0.4)  # Ensure subscription is set up
    mock_get_client.assert_called_once_with(MockArgs())
    mock_client.subscribe.assert_called_once()
    callback = mock_client.subscribe.call_args[0][1]
    callback(mock_client, MQTTPublishPacket(topic="test/topic", payload=b"Test message"))
    callback(mock_client, MQTTPublishPacket(topic="test/topic", payload=b"\x00\x01\x02"))
    callback(mock_client, MQTTPublishPacket(topic="test/topic", payload=b"\xC0\x80"))
    time.sleep(0.1)
    captured = capsys.readouterr()
    out = captured.out.splitlines()
    assert out == [
        "test/topic : Test message",
        "test/topic : <3 bytes of binary data>",
        "test/topic : <2 bytes of binary data>",
    ]
    thread.join(timeout=1.0)



import pytest

from ohmqtt.topic_filter import validate_topic_filter, validate_share_name, match_topic_filter, join_share


def test_topic_filter_empty_filter() -> None:
    f = ""
    with pytest.raises(ValueError):
        validate_topic_filter(f)

def test_topic_filter_empty_topic() -> None:
    f = "sport/tennis/player1"
    with pytest.raises(ValueError):
        match_topic_filter(f, "")

def test_topic_filter_null_character_filter() -> None:
    with pytest.raises(ValueError):
        validate_topic_filter("sport/tennis/\u0000player1")

def test_topic_filter_null_character_topic() -> None:
    f = "sport/tennis/player1"
    with pytest.raises(ValueError):
        match_topic_filter(f, "sport/tennis/\u0000player1")

def test_topic_filter_long_filter() -> None:
    f = "a" * 65536
    with pytest.raises(ValueError):
        validate_topic_filter(f)

def test_topic_filter_long_topic() -> None:
    f = "sport/tennis/player1"
    long_topic = "a" * 65536
    with pytest.raises(ValueError):
        match_topic_filter(f, long_topic)

def test_topic_filter_exact_match() -> None:
    f = "sport/tennis/player1"
    assert match_topic_filter(f, "sport/tennis/player1")
    assert not match_topic_filter(f, "sport/tennis/player2")
    assert not match_topic_filter(f, "sport/tennis")
    assert not match_topic_filter(f, "sport")

def test_topic_filter_multi_level_wildcard_match() -> None:
    f = "sport/tennis/player1/#"
    assert match_topic_filter(f, "sport/tennis/player1")
    assert match_topic_filter(f, "sport/tennis/player1/ranking")
    assert match_topic_filter(f, "sport/tennis/player1/score/wimbledon")
    assert not match_topic_filter(f, "sport/tennis/player2")
    assert not match_topic_filter(f, "sport/tennis")
    assert not match_topic_filter(f, "sport")

def test_topic_filter_multi_level_wildcard_match_hidden() -> None:
    f = "$SYS/#"
    assert match_topic_filter(f, "$SYS/monitor/Clients")

def test_topic_filter_multi_level_wildcard_match_all() -> None:
    f = "#"
    assert match_topic_filter(f, "sport/tennis/player1")
    assert match_topic_filter(f, "sport/tennis/player1/ranking")
    assert match_topic_filter(f, "sport/tennis/player1/score/wimbledon")
    assert match_topic_filter(f, "sport/tennis/player2")
    assert match_topic_filter(f, "sport/tennis")
    assert match_topic_filter(f, "sport")
    assert not match_topic_filter(f, "$SYS/monitor/Clients")

def test_topic_filter_multi_level_wildcard_invalid_filter() -> None:
    with pytest.raises(ValueError):
        validate_topic_filter("sport/tennis#")
    with pytest.raises(ValueError):
        validate_topic_filter("sport/tennis/#/ranking")

def test_topic_filter_multi_level_wildcard_invalid_topic() -> None:
    f = "sport/tennis/player1/#"
    with pytest.raises(ValueError):
        match_topic_filter(f, "sport/tennis/player1/#")

def test_topic_filter_single_level_wildcard_match() -> None:
    f = "sport/tennis/+/ranking"
    assert match_topic_filter(f, "sport/tennis/player1/ranking")
    assert match_topic_filter(f, "sport/tennis/player2/ranking")
    assert not match_topic_filter(f, "sport/tennis/player1")
    assert not match_topic_filter(f, "sport/tennis/player1/score")
    assert not match_topic_filter(f, "sport/tennis/player1/score/wimbledon")
    assert not match_topic_filter(f, "sport/tennis/ranking")

def test_topic_filter_single_level_wildcard_match_hidden() -> None:
    f = "$SYS/+/Clients"
    assert match_topic_filter(f, "$SYS/monitor/Clients")

    f = "+/monitor/Clients"
    assert not match_topic_filter(f, "$SYS/monitor/Clients")

def test_topic_filter_validate_share_name() -> None:
    for share_name in ["", "a/b", "a" * 65536, "\u0000", "#", "+"]:
        try:
            validate_share_name(share_name)
        except ValueError:
            pass
        else:
            pytest.fail(f"Expected ValueError for share name: {share_name}")
    for share_name in ["(foo)", "B A R", "baz-23"]:
        validate_share_name(share_name)

def test_topic_filter_join_share() -> None:
    assert join_share("sport/tennis/player1", "foo") == "$share/foo/sport/tennis/player1"
    assert join_share("sport/tennis/player1", None) == "sport/tennis/player1"
    assert join_share("$SYS/monitor/Clients", "foo") == "$share/foo/$SYS/monitor/Clients"
    assert join_share("$SYS/monitor/Clients", None) == "$SYS/monitor/Clients"



from collections import deque
from dataclasses import dataclass, field
from threading import Condition
from typing import Final, Sequence
import weakref

from .base import LostMessageError, Persistence, RenderedPacket
from ..error import MQTTError
from ..handles import PublishHandle
from ..logger import get_logger
from ..mqtt_spec import MAX_PACKET_ID, MQTTQoS
from ..packet import (
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy

logger: Final = get_logger("persistence.in_memory")


@dataclass(match_args=True, slots=True)
class RetainedMessage:
    """Represents a qos>0 message in the session."""
    topic: str
    payload: bytes
    msg_id: int
    qos: MQTTQoS
    retain: bool
    properties: MQTTPublishProps
    dup: bool
    received: bool
    handle: weakref.ReferenceType[PublishHandle]
    alias_policy: AliasPolicy


@dataclass(slots=True)
class InMemoryPersistence(Persistence):
    """Store for retained messages in the session.

    This store is in memory only and is not persistent."""
    _client_id: str = field(default="", init=False)
    _next_msg_id: int = field(default=1, init=False)
    _messages: dict[int, RetainedMessage] = field(init=False, default_factory=dict)
    _packet_id_to_msg_id: dict[int, int] = field(init=False, default_factory=dict)
    _pending: deque[int] = field(init=False, default_factory=deque)
    _received: set[int] = field(init=False, default_factory=set)
    _cond: Condition = field(init=False, default_factory=Condition)

    def __len__(self) -> int:
        return len(self._messages)

    def add(
        self,
        topic: str,
        payload: bytes,
        qos: MQTTQoS,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> PublishHandle:
        if alias_policy == AliasPolicy.ALWAYS:
            raise ValueError("AliasPolicy must not be ALWAYS for retained messages.")
        msg_id = self._next_msg_id
        self._next_msg_id += 1

        handle = PublishHandle(self._cond)
        message = RetainedMessage(
            topic=topic,
            payload=payload,
            msg_id=msg_id,
            qos=qos,
            retain=retain,
            properties=properties,
            dup=False,
            received=False,
            handle=weakref.ref(handle),
            alias_policy=alias_policy,
        )
        self._messages[msg_id] = message
        self._pending.append(msg_id)
        return handle

    def get(self, count: int) -> Sequence[int]:
        return [self._pending[i] for i in range(min(count, len(self._pending)))]

    def ack(self, packet: MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket) -> None:
        packet_id = packet.packet_id
        rc = packet.reason_code
        if packet_id not in self._packet_id_to_msg_id:
            raise ValueError(f"Unknown packet_id: {packet_id}")
        msg_id = self._packet_id_to_msg_id[packet_id]
        message = self._messages[msg_id]
        if message.qos == MQTTQoS.Q1 or message.received or rc.is_error():
            handle = message.handle()
            if handle is not None:
                with self._cond:
                    handle.ack = packet
                    if rc.is_error():
                        handle.exc = MQTTError("Error code in acknowledgement packet", rc)
                    self._cond.notify_all()
            del self._packet_id_to_msg_id[packet_id]
            del self._messages[msg_id]
        else:
            # Prioritize PUBREL over PUBLISH
            self._pending.appendleft(msg_id)
            message.received = True

    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        if packet.packet_id in self._received:
            logger.debug("Received duplicate QoS 2 packet with ID %d", packet.packet_id)
            return False
        return True

    def set_rec(self, packet: MQTTPublishPacket) -> None:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        self._received.add(packet.packet_id)

    def rel(self, packet: MQTTPubRelPacket) -> None:
        self._received.remove(packet.packet_id)

    def render(self, msg_id: int) -> RenderedPacket:
        packet: MQTTPublishPacket | MQTTPubRelPacket
        msg = self._messages[msg_id]
        packet_id = msg.msg_id
        while packet_id > MAX_PACKET_ID:
            packet_id -= MAX_PACKET_ID
        if msg.received:
            alias_policy = AliasPolicy.NEVER
            packet = MQTTPubRelPacket(packet_id=packet_id)
        else:
            alias_policy = msg.alias_policy
            packet = MQTTPublishPacket(
                topic=msg.topic,
                payload=msg.payload,
                packet_id=packet_id,
                qos=msg.qos,
                retain=msg.retain,
                properties=msg.properties,
                dup=msg.dup,
            )
        if self._pending[0] != msg.msg_id:
            raise ValueError(f"Message {packet_id} is not next in queue.")
        self._pending.popleft()
        self._packet_id_to_msg_id[packet_id] = msg.msg_id
        return RenderedPacket(packet, alias_policy)

    def _reset_inflight(self) -> None:
        """Clear inflight status of all messages."""
        inflight = [i for i in self._messages.keys() if i not in self._pending]
        for msg_id in reversed(inflight):
            self._messages[msg_id].dup = True
            self._pending.appendleft(msg_id)

    def clear(self) -> None:
        with self._cond:
            if self._messages:
                for message in self._messages.values():
                    handle = message.handle()
                    if handle is not None:
                        handle.exc = LostMessageError("Message lost from persistence store")
                self._cond.notify_all()
            self._messages.clear()
            self._pending.clear()
            self._packet_id_to_msg_id.clear()
            self._next_msg_id = 1

    def open(self, client_id: str, clear: bool = False) -> None:
        if clear or client_id != self._client_id:
            self.clear()
            self._client_id = client_id
        else:
            self._reset_inflight()

    def close(self) -> None:
        self.clear()



from __future__ import annotations

from typing import Final, TYPE_CHECKING

from .base import FSMState
from .connecting import ConnectingState
from ..types import ConnectParams, StateData, StateEnvironment
from ...logger import get_logger

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.reconnect_wait")


class ReconnectWaitState(FSMState):
    """Waiting to reconnect to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.timeout.interval = params.reconnect_delay
        state_data.timeout.mark()

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Here we repurpose the keepalive timer to wait for a reconnect.
        if state_data.timeout.exceeded():
            logger.debug("Reconnecting")
            fsm.change_state(ConnectingState)
            return True
        with fsm.cond:
            if fsm.state is not ReconnectWaitState:
                # The state has changed, don't wait.
                return True
            if max_wait is None or max_wait > 0.0:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.cond.wait(timeout)
        return False



import pytest

from ohmqtt.connection.wslib import (
    generate_nonce,
    generate_handshake_key,
    generate_mask,
    apply_mask,
    frame_ws_data,
    deframe_ws_data,
    OpCode,
    WebsocketError,
)


def test_wslib_generate_nonce() -> None:
    nonce1 = generate_nonce()
    nonce2 = generate_nonce()
    assert isinstance(nonce1, str)
    assert isinstance(nonce2, str)
    assert nonce1 != nonce2
    assert len(nonce1) == 24  # 16 bytes base64-encoded is 24 characters
    assert len(nonce2) == 24


def test_wslib_validate_handshake_key() -> None:
    nonce = "dGhlIHNhbXBsZSBub25jZQ=="
    accept_key = "s3pPLMBiTxaQ9kYGzzhZRbK+xOo="
    assert generate_handshake_key(nonce) == accept_key


def test_wslib_generate_mask() -> None:
    mask1 = generate_mask()
    mask2 = generate_mask()
    assert isinstance(mask1, bytes)
    assert isinstance(mask2, bytes)
    assert len(mask1) == 4
    assert len(mask2) == 4
    assert mask1 != mask2


def test_wslib_apply_mask() -> None:
    mask = b"\xFF\x00\xFF\x00"
    data = b"Hello, WebSocket!"
    masked_data = apply_mask(mask, data)
    unmasked_data = apply_mask(mask, masked_data)
    assert masked_data != data
    assert unmasked_data == data


@pytest.mark.parametrize("data_length", [0, 1, 2, 3, 4, 5])
def test_wslib_apply_mask_varied_lengths(data_length: int) -> None:
    mask = b"\xFF\x00\xFF\x00"
    data = bytes(range(data_length))
    masked_data = apply_mask(mask, data)
    unmasked_data = apply_mask(mask, masked_data)
    if data_length > 0:
        assert masked_data != data
    assert unmasked_data == data


@pytest.mark.parametrize("opcode", [OpCode.CONT, OpCode.TEXT, OpCode.BINARY, OpCode.CLOSE, OpCode.PING, OpCode.PONG])
@pytest.mark.parametrize("data_length", [0, 125, 126, 127, 65535, 65536])
@pytest.mark.parametrize("do_mask", [True, False])
def test_wslib_frame_ws_data(opcode: OpCode, data_length: int, do_mask: bool) -> None:
    data = b"\x88" * data_length
    framed_data = frame_ws_data(opcode, data, do_mask=do_mask)

    assert framed_data[0] == 0x80 | opcode.value  # FIN + opcode
    length_byte = framed_data[1]
    mask_bit = 0x80 if do_mask else 0x00
    assert (length_byte & 0x80) == mask_bit
    if data_length <= 125:
        assert (length_byte & 0x7F) == data_length
        index = 2
    elif data_length <= 0xFFFF:
        assert (length_byte & 0x7F) == 126
        extended_length = int.from_bytes(framed_data[2:4], "big")
        assert extended_length == data_length
        index = 4
    else:
        assert (length_byte & 0x7F) == 127
        extended_length = int.from_bytes(framed_data[2:10], "big")
        assert extended_length == data_length
        index = 10
    if do_mask:
        mask = framed_data[index:index+4]
        masked_payload = framed_data[index+4:]
        payload = apply_mask(mask, masked_payload)
    else:
        payload = framed_data[index:]
    assert payload == data


@pytest.mark.parametrize("opcode", [OpCode.CONT, OpCode.TEXT, OpCode.BINARY, OpCode.CLOSE, OpCode.PING, OpCode.PONG])
@pytest.mark.parametrize("data_length", [0, 125, 126, 127, 65535, 65536])
@pytest.mark.parametrize("do_mask", [True, False])
def test_wslib_deframe_ws_data(opcode: OpCode, data_length: int, do_mask: bool) -> None:
    data = b"\x88" * data_length
    framed_data = frame_ws_data(opcode, data, do_mask=do_mask)

    frame = deframe_ws_data(bytearray(framed_data))
    assert frame is not None
    opcode, payload, was_masked, length = frame
    assert opcode == opcode
    assert payload == data
    assert was_masked == do_mask
    assert length == len(framed_data)


@pytest.mark.parametrize("opcode", [OpCode.CONT, OpCode.TEXT, OpCode.BINARY, OpCode.CLOSE, OpCode.PING, OpCode.PONG])
@pytest.mark.parametrize("data_length", [0, 125, 126, 127, 65535, 65536])
@pytest.mark.parametrize("do_mask", [True, False])
def test_wslib_deframe_ws_data_drip_feed(opcode: OpCode, data_length: int, do_mask: bool) -> None:
    data = b"\x88" * data_length
    framed_data = frame_ws_data(opcode, data, do_mask=do_mask)

    # Only drip feed enough for confidence that we've covered all the decoding branches.
    feed_length = min(len(framed_data) - 1, 16)
    for i in range(feed_length):
        result = deframe_ws_data(bytearray(framed_data[:i]))
        assert result is None


def test_wslib_deframe_ws_data_fragment() -> None:
    framed_data = bytearray([OpCode.BINARY, 0x00])  # FIN=0, opcode=BINARY, no payload
    with pytest.raises(WebsocketError, match="Fragmented WebSocket frames are not supported"):
        deframe_ws_data(framed_data)


@pytest.mark.parametrize("do_mask", [True, False])
def test_wslib_frame_large_data(do_mask: bool) -> None:
    data_length = 256 * 1024 * 1024  # 256MB, maximum MQTT packet size
    data = b"\x88" * data_length
    framed_data = frame_ws_data(OpCode.BINARY, data, do_mask=do_mask)
    frame = deframe_ws_data(bytearray(framed_data))
    assert frame is not None
    opcode, payload, was_masked, length = frame
    assert opcode == OpCode.BINARY
    assert payload == data
    assert was_masked is do_mask
    assert length == len(framed_data)



from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection.timeout import Timeout


@pytest.fixture
def mock_time(mocker: MockerFixture) -> Mock:
    return mocker.patch("ohmqtt.connection.timeout._time")


def test_timeout(mock_time: Mock) -> None:
    mock_time.return_value = 0.0
    timeout = Timeout()

    assert timeout.interval is None
    assert timeout.get_timeout() is None
    assert not timeout.exceeded()

    timeout.interval = 1.0
    assert timeout.get_timeout() == 1.0
    assert not timeout.exceeded()

    mock_time.return_value = 1.0
    assert timeout.get_timeout() == 0.0
    assert timeout.exceeded()

    mock_time.return_value = 2.0
    assert timeout.get_timeout() == 0.0
    assert timeout.exceeded()

    timeout.mark()
    assert timeout.get_timeout() == 1.0
    assert not timeout.exceeded()


def test_timeout_max_wait(mock_time: Mock) -> None:
    mock_time.return_value = 0.0
    timeout = Timeout()

    assert timeout.interval is None
    assert timeout.get_timeout(max_wait=1.0) == 1.0

    timeout.interval = 1.0
    assert timeout.get_timeout(max_wait=0.5) == 0.5



"""
The Keep Alive is a Two Byte Integer (a required field in CONNECT packets)
which is a time interval measured in seconds.

It is the maximum time interval that is permitted to elapse between the point
at which the Client finishes transmitting one MQTT Control Packet and the
point it starts sending the next.

It is the responsibility of the Client to ensure that the interval between
MQTT Control Packets being sent does not exceed the Keep Alive value.

If Keep Alive is non-zero and in the absence of sending any other MQTT Control
Packets, the Client MUST send a PINGREQ packet.

If the Server returns a Server Keep Alive on the CONNACK packet, the Client
MUST use that value instead of the value it sent as the Keep Alive.

If the Server does not send the Server Keep Alive, the Server MUST use the
Keep Alive value set by the Client on CONNECT.

If the Keep Alive value is non-zero and the Server does not receive an MQTT
Control Packet from the Client within one and a half times the Keep Alive time
period, it MUST close the Network Connection to the Client as if the network
had failed.

If a Client does not receive a PINGRESP packet within a reasonable amount of
time after it has sent a PINGREQ, it SHOULD close the Network Connection to
the Server.

A Keep Alive value of 0 has the effect of turning off the Keep Alive
mechanism. If Keep Alive is 0 the Client is not obliged to send MQTT Control
Packets on any particular schedule.
"""

from time import monotonic as _time
from typing import Final


MIN_TIMEOUT: Final = 0


class KeepAlive:
    """Tracks the keep alive timer for a connection."""
    __slots__ = (
        "_keepalive_interval",
        "_last_send",
        "_pong_deadline",
    )

    def __init__(self) -> None:
        self._keepalive_interval = 0
        self._last_send = 0.0
        self._pong_deadline = 0.0

    @property
    def keepalive_interval(self) -> int:
        """Get the keep alive interval in seconds."""
        return self._keepalive_interval

    @keepalive_interval.setter
    def keepalive_interval(self, value: int) -> None:
        """Set the keep alive interval in seconds.

        :raises ValueError: The value is out of range."""
        if value < 0:
            raise ValueError("Keep alive interval must be non-negative")
        if value > 65535:
            raise ValueError("Keep alive interval must be <= 65535")
        if value != self._keepalive_interval:
            self._keepalive_interval = value
            # Reset PINGRESP expectation.
            self._pong_deadline = 0.0

    def should_close(self) -> bool:
        """Check if the keep alive timer has expired."""
        if self._keepalive_interval == 0:
            # Keep alive is disabled
            return False
        now = _time()
        if self._pong_deadline > 0.0 and now >= self._pong_deadline:
            # No PINGRESP received after sending PINGREQ
            return True
        return False

    def should_send_ping(self) -> bool:
        """Check if a PINGREQ packet should be sent."""
        if self._keepalive_interval == 0:
            # Keep alive is disabled
            return False
        now = _time()
        if now - self._last_send >= self._keepalive_interval and self._pong_deadline == 0.0:
            # No data sent for keepalive_interval and no PINGREQ inflight
            return True
        return False

    def get_next_timeout(self, max_wait: float | None = None) -> float | None:
        """Check how long until the next next ping or closure check is due.

        max_wait is the maximum time to wait for a timeout. If this is set,
        the timeout will be the minimum of the calculated timeout and max_wait.

        If the keep alive interval is 0, this will return None unless max_wait is set."""
        if self._keepalive_interval == 0:
            # Keep alive is disabled
            if max_wait is not None:
                # If max_wait is set, we should wait for that time.
                return max_wait
            return None
        now = _time()
        if self._pong_deadline > 0.0:
            # After this amount of time, we should close the connection.
            to = max(MIN_TIMEOUT, self._pong_deadline - now)
        else:
            # After this amount of time, we should send a PINGREQ.
            to = max(MIN_TIMEOUT, self._last_send + self._keepalive_interval - now)
        if max_wait is not None:
            to = min(to, max_wait)
        return to

    def mark_init(self) -> None:
        """Initialize the keep alive timer for a new connection."""
        self._last_send = _time()
        self._pong_deadline = 0.0

    def mark_send(self) -> None:
        """Mark that data has been sent."""
        self._last_send = _time()

    def mark_ping(self) -> None:
        """Mark that a PINGREQ was sent.

        This also marks that data has been sent."""
        now = _time()
        self._pong_deadline = now + self._keepalive_interval
        self._last_send = now

    def mark_pong(self) -> None:
        """Mark that a PINGRESP was received.

        This also marks that data has been received."""
        self._pong_deadline = 0.0



"""
Fake MQTT broker for testing purposes.
"""

from __future__ import annotations

from collections import deque
import socket
import socketserver
import threading
import time
from typing import Callable, cast, Final
import uuid

from ohmqtt.connection.decoder import ClosedSocketError, IncrementalDecoder
from ohmqtt.connection.wslib import deframe_ws_data, frame_ws_data, generate_handshake_key, OpCode
from ohmqtt.logger import get_logger
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import (
    decode_packet,
    MQTTPacket,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    MQTTDisconnectPacket,
)
from ohmqtt.property import MQTTConnAckProps
from ohmqtt.topic_filter import match_topic_filter, validate_topic_filter

logger: Final = get_logger("tests.util.fake_broker")


class FakeBrokerHandler(socketserver.BaseRequestHandler):
    server: FakeBrokerServer
    subscriptions: set[str]

    def setup(self) -> None:
        self.decoder = IncrementalDecoder()
        self.subscriptions: set[str] = set()

    def handle(self) -> None:
        self.server.sock = self.request
        try:
            while (packet := self.decoder.decode(self.request)) is not None:
                self._handle_packet(packet)
        except ClosedSocketError:
            pass

    def handle_error(self, request: socket.socket, client_address: tuple[str, int]) -> None:
        logger.exception("Exception in handler for %s", client_address)

    def _handle_packet(self, packet: MQTTPacket) -> None:
        outbound: list[MQTTPacket] = []
        logger.info("FakeBroker <--- %s", packet)
        self.server.received.append(packet)

        handlers = {
            MQTTConnectPacket: self._handle_connect,
            MQTTSubscribePacket: self._handle_subscribe,
            MQTTUnsubscribePacket: self._handle_unsubscribe,
            MQTTPublishPacket: self._handle_publish,
            MQTTPubRecPacket: self._handle_pubrec,
            MQTTPubRelPacket: self._handle_pubrel,
            MQTTPingReqPacket: self._handle_pingreq,
            MQTTDisconnectPacket: self._handle_disconnect,
        }

        if type(packet) in handlers:
            handler = cast(Callable[[MQTTPacket], list[MQTTPacket]], handlers[type(packet)])
            outbound.extend(handler(packet))

        for pkt in outbound:
            logger.info("FakeBroker ---> %s", pkt)
            self._send_packet(pkt)

    def _send_packet(self, packet: MQTTPacket) -> None:
        self.request.sendall(packet.encode())

    def _handle_connect(self, packet: MQTTConnectPacket) -> list[MQTTPacket]:
        if packet.client_id:
            connack = MQTTConnAckPacket()
        else:
            client_id = f"auto-{uuid.uuid4()!s}"
            connack = MQTTConnAckPacket(
                properties=MQTTConnAckProps(AssignedClientIdentifier=client_id)
            )
        return [connack]

    def _handle_subscribe(self, packet: MQTTSubscribePacket) -> list[MQTTPacket]:
        for topic_filter, opts in packet.topics:
            validate_topic_filter(topic_filter)
            if opts & 0x04:
                # no_local is set, do not mirror.
                continue
            self.subscriptions.add(topic_filter)
        return [MQTTSubAckPacket(packet_id=packet.packet_id, reason_codes=[MQTTReasonCode.Success] * len(packet.topics))]

    def _handle_unsubscribe(self, packet: MQTTUnsubscribePacket) -> list[MQTTPacket]:
        for topic in packet.topics:
            validate_topic_filter(topic)
            self.subscriptions.discard(topic)
        return [MQTTUnsubAckPacket(packet_id=packet.packet_id, reason_codes=[MQTTReasonCode.Success] * len(packet.topics))]

    def _handle_publish(self, packet: MQTTPublishPacket) -> list[MQTTPacket]:
        outbound: list[MQTTPacket] = []
        if any(match_topic_filter(topic_filter, packet.topic) for topic_filter in self.subscriptions):
            outbound.append(packet)

        if packet.qos == MQTTQoS.Q1:
            outbound.append(MQTTPubAckPacket(packet_id=packet.packet_id))
        elif packet.qos == MQTTQoS.Q2:
            outbound.append(MQTTPubRecPacket(packet_id=packet.packet_id))

        return outbound

    def _handle_pubrec(self, packet: MQTTPubRecPacket) -> list[MQTTPacket]:
        return [MQTTPubRelPacket(packet_id=packet.packet_id)]

    def _handle_pubrel(self, packet: MQTTPubRelPacket) -> list[MQTTPacket]:
        return [MQTTPubCompPacket(packet_id=packet.packet_id)]

    def _handle_pingreq(self, packet: MQTTPingReqPacket) -> list[MQTTPacket]:
        return [MQTTPingRespPacket()]

    def _handle_disconnect(self, packet: MQTTDisconnectPacket) -> list[MQTTPacket]:
        self.request.close()
        return []


class FakeWebsocketBrokerHandler(FakeBrokerHandler):
    buffer: bytearray
    handshake_done: bool

    def setup(self) -> None:
        self.buffer = bytearray()
        self.handshake_done = False
        self.subscriptions: set[str] = set()

    def handle(self) -> None:
        logger.debug("FakeWebsocketBrokerHandler handle()")
        if not self.handshake_done:
            self.server.sock = self.request
            data = self.request.recv(0xffff)
            if not data:
                return  # Connection closed
            lines = data.split(b"\r\n")
            head = lines[0]
            assert head.startswith(b"GET "), "Invalid websocket handshake method"
            assert head.endswith(b" HTTP/1.1"), "Invalid websocket handshake version"
            headers = {
                line.split(b":", 1)[0].lower().strip(): line.split(b":", 1)[1].strip()
                for line in lines[1:]
                if b": " in line
            }
            assert b"sec-websocket-key" in headers, "Invalid websocket handshake"
            assert headers.get(b"upgrade", b"").lower() == b"websocket", "Invalid websocket upgrade header"
            assert headers.get(b"connection", b"").lower() == b"upgrade", "Invalid websocket connection header"
            assert headers.get(b"sec-websocket-version", b"") == b"13", "Invalid websocket version header"
            assert headers.get(b"sec-websocket-protocol", b"") == b"mqtt", "Invalid websocket protocol header"
            nonce = headers[b"sec-websocket-key"]
            accept_key = generate_handshake_key(nonce.decode("utf-8"))

            response = (
                "HTTP/1.1 101 Switching Protocols\r\n"
                "Upgrade: websocket\r\n"
                "Connection: Upgrade\r\n"
                f"Sec-WebSocket-Accept: {accept_key}\r\n"
                "Sec-WebSocket-Protocol: mqtt\r\n"
                "\r\n"
            ).encode("utf-8")
            self.request.sendall(response)
            self.handshake_done = True
            logger.info("FakeBroker handshake responded")

        try:
            while (data := self.request.recv(0xffff)):
                self.buffer.extend(data)
                logger.debug("FakeBroker has %d bytes of websocket data", len(self.buffer))

                while (frame := deframe_ws_data(self.buffer)) is not None:
                    opcode, payload, was_masked, frame_length = frame
                    assert was_masked, "Client-to-broker frames must be masked"
                    del self.buffer[:frame_length]
                    if opcode == OpCode.BINARY:
                        packet = decode_packet(payload)
                        self._handle_packet(packet)
        except OSError:
            pass

    def _send_packet(self, packet: MQTTPacket) -> None:
        frame = frame_ws_data(OpCode.BINARY, packet.encode(), do_mask=False)
        self.request.sendall(frame)


class FakeBrokerServer(socketserver.TCPServer):
    received: deque[MQTTPacket]
    sock: socket.socket | None

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.received = deque()
        self.sock = None


class FakeBroker(threading.Thread):
    handler_class: type[socketserver.BaseRequestHandler] = FakeBrokerHandler

    def __init__(self) -> None:
        super().__init__(daemon=True)
        self.server = FakeBrokerServer(("localhost", 0), self.handler_class)
        self.port = self.server.server_address[1]

    def __enter__(self) -> FakeBroker:
        self.start()
        ready = False
        t0 = time.monotonic()
        while not ready and time.monotonic() - t0 < 5.0:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.connect(("localhost", self.port))
                ready = True
            except ConnectionRefusedError:
                time.sleep(0.1)
        return self

    def __exit__(self, *args: object) -> None:
        # Shutting down the server deadlocks here, so pass and let it be dereferenced.
        pass

    @property
    def received(self) -> deque[MQTTPacket]:
        return self.server.received

    @property
    def sock(self) -> socket.socket | None:
        return self.server.sock

    def run(self) -> None:
        with self.server:
            self.server.serve_forever()

    def shutdown(self) -> None:
        self.server.shutdown()
        self.server.server_close()

    def send(self, packet: MQTTPacket) -> None:
        """Send a packet from the broker to the client."""
        if self.sock is None:
            raise RuntimeError("Broker-to-client socket is not initialized")
        logger.info("FakeBroker ---> %s", packet)
        self.sock.sendall(packet.encode())


class FakeWebsocketBroker(FakeBroker):
    handler_class: type[socketserver.BaseRequestHandler] = FakeWebsocketBrokerHandler

    def send(self, packet: MQTTPacket) -> None:
        """Send a packet from the broker to the client."""
        if self.sock is None:
            raise RuntimeError("Broker-to-client socket is not initialized")
        logger.info("FakeBroker ---> %s", packet)
        frame = frame_ws_data(OpCode.BINARY, packet.encode(), do_mask=False)
        self.sock.sendall(frame)



from dataclasses import dataclass, field
import socket
import ssl
from typing import NamedTuple, Final

from .decoder import ClosedSocketError, WantReadError
from .wslib import OpCode, WebsocketError
from ..logger import get_logger

logger: Final = get_logger("connection.ws_decoder")


class LengthDecodeResult(NamedTuple):
    """Result of decoding frame length, in part or whole.

    This state can be used to resume decoding if the socket doesn't have enough data."""
    value: int
    remaining: int


InitLengthDecodeState: Final = LengthDecodeResult(0, 9)


@dataclass(slots=True)
class WebsocketDecoder:
    """Incremental decoder for Websocket frames coming in from a socket."""
    opcode: OpCode | None = field(default=None, init=False)  #: The opcode of the frame.
    length: LengthDecodeResult = field(default=InitLengthDecodeState, init=False)  #: Variable integer decoding state of the payload length.
    data: bytearray = field(init=False, default_factory=bytearray)  #: The payload of the frame.

    def reset(self) -> None:
        """Reset the decoder state."""
        self.opcode = None
        self.length = InitLengthDecodeState
        self.data.clear()

    def _recv_one_byte(self, sock: socket.socket | ssl.SSLSocket) -> int:
        """Receive one byte from the socket.

        :raises WantReadError: The socket is not ready for reading.
        :raises ClosedSocketError: The socket is closed."""
        try:
            data = sock.recv(1)
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            raise WantReadError("Socket not ready for reading")
        if not data:
            raise ClosedSocketError("Socket closed")
        return data[0]

    def _extract_opcode(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Extract the opcode from a WebSocket frame, if needed.

        :raises WantReadError: The socket is not ready for reading.
        :raises WebsocketError: The frame is invalid."""
        if self.opcode is not None:
            return
        byte = self._recv_one_byte(sock)
        if byte & 0x80 == 0:
            raise WebsocketError("Fragmented WebSocket frames are not supported")
        if byte & 0x70 != 0:
            raise WebsocketError("Reserved WebSocket bits are set")
        opcode_value = byte & 0x0F
        try:
            self.opcode = OpCode(opcode_value)
        except ValueError as exc:
            raise WebsocketError("Invalid WebSocket opcode") from exc

    def _extract_length(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Incrementally decode frame length integer from a socket, if needed.

        :raises WantReadError: The socket is not ready for reading.
        :raises WebsocketError: The frame is invalid."""
        result = self.length.value
        remaining = self.length.remaining
        if remaining == 0:
            return

        try:
            if remaining > 8:
                # First byte of length.
                byte = self._recv_one_byte(sock)
                if byte & 0x80:
                    raise WebsocketError("Masked frames from server are not allowed")
                if byte <= 125:
                    result = byte
                    remaining = 0
                elif byte == 126:
                    remaining = 2
                else:
                    remaining = 8

            while remaining > 0:
                byte = self._recv_one_byte(sock)
                result = (result << 8) + byte
                remaining -= 1

            # We have the complete length.
            self.length = LengthDecodeResult(result, 0)
        except WantReadError:
            # Not done yet, the socket is neither closed nor ready for reading.
            # Save the partial state and re-raise.
            self.length = LengthDecodeResult(result, remaining)
            raise

    def _extract_data(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Extract all data after the packet length from the socket, if needed.

        :raises ClosedSocketError: The socket is closed."""
        assert self.length.remaining == 0, "Can't extract data without complete length"
        while len(self.data) < self.length.value:
            data = sock.recv(self.length.value - len(self.data))
            if not data:
                raise ClosedSocketError("Socket closed")
            self.data.extend(data)

    def decode(self, sock: socket.socket | ssl.SSLSocket) -> tuple[OpCode, bytes] | None:
        """Decode a single frame straight from the socket.

        :return: None if the socket doesn't have enough data for us to decode a packet.
        :raises ClosedSocketError: The socket is closed.
        :raises WebsocketError: A protocol error was detected."""
        try:
            self._extract_opcode(sock)
            self._extract_length(sock)
            self._extract_data(sock)
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError, WantReadError):
            # If the socket is open but doesn't have enough data for us, we need to wait for more.
            return None
        except OSError as exc:
            raise ClosedSocketError("Socket closed") from exc

        assert self.opcode is not None
        opcode = self.opcode
        data = bytes(self.data)
        self.reset()
        return opcode, data



from __future__ import annotations

import sys
from types import SimpleNamespace
from typing import Callable, cast, Final, Mapping, NamedTuple, Sequence, TypeAlias

if sys.version_info >= (3, 11):
    from typing import dataclass_transform, Self
else:
    from typing_extensions import dataclass_transform, Self

from .error import MQTTError
from .mqtt_spec import MQTTPropertyId, MQTTReasonCode
from .serialization import (
    encode_bool,
    decode_bool,
    encode_uint8,
    decode_uint8,
    encode_uint16,
    decode_uint16,
    encode_uint32,
    decode_uint32,
    encode_string,
    decode_string,
    encode_string_pair,
    decode_string_pair,
    encode_binary,
    decode_binary,
    encode_varint,
    decode_varint,
)


@dataclass_transform(kw_only_default=True)
class MQTTPropertiesBase(SimpleNamespace):
    """Represents MQTT packet properties."""
    def __len__(self) -> int:
        return len(self.__dict__)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(f'{key}={value}' for key, value in self.__dict__.items())})"


BoolFieldT: TypeAlias = bool
IntFieldT: TypeAlias = int
StrFieldT: TypeAlias = str
BytesFieldT: TypeAlias = bytes
SetIntFieldT: TypeAlias = set[int]
SeqStrPairFieldT: TypeAlias = Sequence[tuple[str, str]]
AnyFieldT: TypeAlias = IntFieldT | StrFieldT | BytesFieldT | BoolFieldT | SetIntFieldT | SeqStrPairFieldT

PayloadFormatIndicatorT: TypeAlias = IntFieldT | None
MessageExpiryIntervalT: TypeAlias = IntFieldT | None
ContentTypeT: TypeAlias = StrFieldT | None
ResponseTopicT: TypeAlias = StrFieldT | None
CorrelationDataT: TypeAlias = BytesFieldT | None
SubscriptionIdentifierT: TypeAlias = SetIntFieldT | None
TopicAliasT: TypeAlias = IntFieldT | None
SessionExpiryIntervalT: TypeAlias = IntFieldT | None
AssignedClientIdentifierT: TypeAlias = StrFieldT | None
ServerKeepAliveT: TypeAlias = IntFieldT | None
AuthenticationMethodT: TypeAlias = StrFieldT | None
AuthenticationDataT: TypeAlias = BytesFieldT | None
RequestProblemInformationT: TypeAlias = BoolFieldT | None
WillDelayIntervalT: TypeAlias = IntFieldT | None
RequestResponseInformationT: TypeAlias = BoolFieldT | None
ResponseInformationT: TypeAlias = StrFieldT | None
ServerReferenceT: TypeAlias = StrFieldT | None
ReasonStringT: TypeAlias = StrFieldT | None
ReceiveMaximumT: TypeAlias = IntFieldT | None
TopicAliasMaximumT: TypeAlias = IntFieldT | None
MaximumQoST: TypeAlias = IntFieldT | None
RetainAvailableT: TypeAlias = BoolFieldT | None
UserPropertyT: TypeAlias = SeqStrPairFieldT | None
MaximumPacketSizeT: TypeAlias = IntFieldT | None
WildcardSubscriptionAvailableT: TypeAlias = BoolFieldT | None
SubscriptionIdentifierAvailableT: TypeAlias = BoolFieldT | None
SharedSubscriptionAvailableT: TypeAlias = BoolFieldT | None


# Build some data structures for fast serialization/deserialization of properties.
_SerializerTypes: TypeAlias = (
    Callable[[bool], bytes] |
    Callable[[int], bytes] |
    Callable[[str], bytes] |
    Callable[[bytes], bytes] |
    Callable[[tuple[str, str]], bytes]
)
_DeserializerTypes: TypeAlias = (
    Callable[[memoryview], tuple[bool, int]] |
    Callable[[memoryview], tuple[int, int]] |
    Callable[[memoryview], tuple[str, int]] |
    Callable[[memoryview], tuple[bytes, int]] |
    Callable[[memoryview], tuple[tuple[str, str], int]]
)
class _SerializerPair(NamedTuple):
    """Pair of serializer and deserializer functions for a property."""
    serializer: _SerializerTypes
    deserializer: _DeserializerTypes
_BoolSerializer: Final = _SerializerPair(serializer=encode_bool, deserializer=decode_bool)
_UInt8Serializer: Final = _SerializerPair(serializer=encode_uint8, deserializer=decode_uint8)
_UInt16Serializer: Final = _SerializerPair(serializer=encode_uint16, deserializer=decode_uint16)
_UInt32Serializer: Final = _SerializerPair(serializer=encode_uint32, deserializer=decode_uint32)
_VarIntSerializer: Final = _SerializerPair(serializer=encode_varint, deserializer=decode_varint)
_BinarySerializer: Final = _SerializerPair(serializer=encode_binary, deserializer=decode_binary)
_StringSerializer: Final = _SerializerPair(serializer=encode_string, deserializer=decode_string)
_StringPairSerializer: Final = _SerializerPair(serializer=encode_string_pair, deserializer=decode_string_pair)

_PropertySerializers: Final[Mapping[str, _SerializerPair]] = {
    MQTTPropertyId.PayloadFormatIndicator.name: _UInt8Serializer,
    MQTTPropertyId.MessageExpiryInterval.name: _UInt32Serializer,
    MQTTPropertyId.ContentType.name: _StringSerializer,
    MQTTPropertyId.ResponseTopic.name: _StringSerializer,
    MQTTPropertyId.CorrelationData.name: _BinarySerializer,
    MQTTPropertyId.SubscriptionIdentifier.name: _VarIntSerializer,  # This serializer is used for each set element.
    MQTTPropertyId.SessionExpiryInterval.name: _UInt32Serializer,
    MQTTPropertyId.AssignedClientIdentifier.name: _StringSerializer,
    MQTTPropertyId.ServerKeepAlive.name: _UInt16Serializer,
    MQTTPropertyId.AuthenticationMethod.name: _StringSerializer,
    MQTTPropertyId.AuthenticationData.name: _BinarySerializer,
    MQTTPropertyId.RequestProblemInformation.name: _BoolSerializer,
    MQTTPropertyId.WillDelayInterval.name: _UInt32Serializer,
    MQTTPropertyId.RequestResponseInformation.name: _BoolSerializer,
    MQTTPropertyId.ResponseInformation.name: _StringSerializer,
    MQTTPropertyId.ServerReference.name: _StringSerializer,
    MQTTPropertyId.ReasonString.name: _StringSerializer,
    MQTTPropertyId.ReceiveMaximum.name: _UInt16Serializer,
    MQTTPropertyId.TopicAliasMaximum.name: _UInt16Serializer,
    MQTTPropertyId.TopicAlias.name: _UInt16Serializer,
    MQTTPropertyId.MaximumQoS.name: _UInt8Serializer,
    MQTTPropertyId.RetainAvailable.name: _BoolSerializer,
    MQTTPropertyId.UserProperty.name: _StringPairSerializer,  # This serializer is used for each Sequence element.
    MQTTPropertyId.MaximumPacketSize.name: _UInt32Serializer,
    MQTTPropertyId.WildcardSubscriptionAvailable.name: _BoolSerializer,
    MQTTPropertyId.SubscriptionIdentifierAvailable.name: _BoolSerializer,
    MQTTPropertyId.SharedSubscriptionAvailable.name: _BoolSerializer
}
_PropertySerializersById: Final[Mapping[int, _SerializerPair]] = {
    getattr(MQTTPropertyId, key): value for key, value in _PropertySerializers.items()
}


class MQTTProperties(MQTTPropertiesBase):
    def encode(self) -> bytes:
        """Encode MQTT properties to a buffer."""
        properties = self.__dict__
        if not properties:
            # Fast path for empty properties.
            return b"\x00"
        data = bytearray()
        for key, prop_value in properties.items():
            prop_id = getattr(MQTTPropertyId, key)
            serializer = _PropertySerializers[key].serializer

            # MQTT specification calls for a variable integer for the property ID,
            #   but we know that the IDs are all 1 byte long,
            #   so we will encode them as single bytes to save a branch.

            if key in ("SubscriptionIdentifier", "UserProperty"):
                # These properties may appear multiple times, in any order.
                for sub_value in prop_value:
                    data.append(prop_id)
                    data.extend(serializer(sub_value))
            else:
                data.append(prop_id)
                data.extend(serializer(prop_value))
        data[0:0] = encode_varint(len(data))
        return bytes(data)

    @classmethod
    def decode(cls, data: memoryview) -> tuple[Self, int]:
        """Decode MQTT properties from a buffer.

        Returns a tuple of the decoded properties and the number of bytes consumed."""
        length, length_sz = decode_varint(data)
        if length == 0:
            # Fast path for empty properties.
            return cls(), 1
        data = data[length_sz:]
        remaining = length
        properties: dict[str, AnyFieldT] = {}
        while remaining:
            # The spec calls for a variable integer for the property ID,
            #   but we know that the IDs are all 1 byte long,
            #   so we will decode them as uint8 to save a branch.
            key = data[0]
            data  = data[1:]
            remaining -= 1
            prop_name = MQTTPropertyId(key).name
            deserializer = _PropertySerializersById[key].deserializer
            value, sz = deserializer(data)
            data = data[sz:]
            remaining -= sz
            if prop_name == "SubscriptionIdentifier":
                # This property may appear multiple times, in any order.
                if prop_name not in properties:
                    properties[prop_name] = set()
                cast(SetIntFieldT, properties[prop_name]).add(cast(int, value))
            elif prop_name == "UserProperty":
                # This property may appear multiple times, in any order.
                if prop_name not in properties:
                    properties[prop_name] = []
                cast(list[tuple[str, str]], properties[prop_name]).append(cast(tuple[str, str], value))
            elif prop_name in properties:
                # Other properties must appear exactly once.
                raise MQTTError(f"Duplicate property {prop_name}", MQTTReasonCode.ProtocolError)
            else:
                properties[prop_name] = value  # type: ignore[assignment]
        return cls(**properties), length + length_sz


class MQTTConnectProps(MQTTProperties):
    """Properties for MQTT CONNECT packet."""
    SessionExpiryInterval: SessionExpiryIntervalT = None
    ReceiveMaximum: ReceiveMaximumT = None
    MaximumPacketSize: MaximumPacketSizeT = None
    TopicAliasMaximum: TopicAliasMaximumT = None
    RequestResponseInformation: RequestResponseInformationT = None
    RequestProblemInformation: RequestProblemInformationT = None
    AuthenticationMethod: AuthenticationMethodT = None
    AuthenticationData: AuthenticationDataT = None
    UserProperty: UserPropertyT = None


class MQTTConnAckProps(MQTTProperties):
    """Properties for MQTT CONNACK packet."""
    SessionExpiryInterval: SessionExpiryIntervalT = None
    ReceiveMaximum: ReceiveMaximumT = None
    MaximumQoS: MaximumQoST = None
    RetainAvailable: RetainAvailableT = None
    MaximumPacketSize: MaximumPacketSizeT = None
    AssignedClientIdentifier: AssignedClientIdentifierT = None
    TopicAliasMaximum: TopicAliasMaximumT = None
    ReasonString: ReasonStringT = None
    WildcardSubscriptionAvailable: WildcardSubscriptionAvailableT = None
    SubscriptionIdentifierAvailable: SubscriptionIdentifierAvailableT = None
    SharedSubscriptionAvailable: SharedSubscriptionAvailableT = None
    ServerKeepAlive: ServerKeepAliveT = None
    ResponseInformation: ResponseInformationT = None
    ServerReference: ServerReferenceT = None
    AuthenticationMethod: AuthenticationMethodT = None
    AuthenticationData: AuthenticationDataT = None
    UserProperty: UserPropertyT = None


class MQTTPublishProps(MQTTProperties):
    """Properties for MQTT PUBLISH packet."""
    PayloadFormatIndicator: PayloadFormatIndicatorT = None
    MessageExpiryInterval: MessageExpiryIntervalT = None
    ContentType: ContentTypeT = None
    ResponseTopic: ResponseTopicT = None
    CorrelationData: CorrelationDataT = None
    SubscriptionIdentifier: SubscriptionIdentifierT = None
    TopicAlias: TopicAliasT = None
    UserProperty: UserPropertyT = None


class MQTTPubAckProps(MQTTProperties):
    """Properties for MQTT PUBACK packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTPubRecProps(MQTTProperties):
    """Properties for MQTT PUBREC packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTPubRelProps(MQTTProperties):
    """Properties for MQTT PUBREL packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTPubCompProps(MQTTProperties):
    """Properties for MQTT PUBCOMP packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTSubscribeProps(MQTTProperties):
    """Properties for MQTT SUBSCRIBE packet."""
    SubscriptionIdentifier: SubscriptionIdentifierT = None
    UserProperty: UserPropertyT = None


class MQTTSubAckProps(MQTTProperties):
    """Properties for MQTT SUBACK packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTUnsubscribeProps(MQTTProperties):
    """Properties for MQTT UNSUBSCRIBE packet."""
    UserProperty: UserPropertyT = None


class MQTTUnsubAckProps(MQTTProperties):
    """Properties for MQTT UNSUBACK packet."""
    ReasonString: ReasonStringT = None
    UserProperty: UserPropertyT = None


class MQTTDisconnectProps(MQTTProperties):
    """Properties for MQTT DISCONNECT packet."""
    SessionExpiryInterval: SessionExpiryIntervalT = None
    ReasonString: ReasonStringT = None
    ServerReference: ServerReferenceT = None
    UserProperty: UserPropertyT = None


class MQTTAuthProps(MQTTProperties):
    """Properties for MQTT AUTH packet."""
    ReasonString: ReasonStringT = None
    AuthenticationMethod: AuthenticationMethodT = None
    AuthenticationData: AuthenticationDataT = None
    UserProperty: UserPropertyT = None


class MQTTWillProps(MQTTProperties):
    """Properties for MQTT Will message."""
    PayloadFormatIndicator: PayloadFormatIndicatorT = None
    MessageExpiryInterval: MessageExpiryIntervalT = None
    ContentType: ContentTypeT = None
    ResponseTopic: ResponseTopicT = None
    CorrelationData: CorrelationDataT = None
    WillDelayInterval: WillDelayIntervalT = None
    UserProperty: UserPropertyT = None



from abc import ABCMeta, abstractmethod
import argparse
from typing import Any, TypeAlias, TYPE_CHECKING

from .. import Client, MQTTConnectProps


if TYPE_CHECKING:
    SubParsersT: TypeAlias = argparse._SubParsersAction[argparse.ArgumentParser]  # noqa: SLF001
else:
    SubParsersT = Any


class BrokerConnectionError(Exception):
    """Failed to connect to the MQTT broker."""


def add_broker_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Add common broker connection arguments to the parser."""
    parser.add_argument(
        "--address", type=str, default="localhost", help="The MQTT broker address"
    )
    parser.add_argument(
        "--client_id", type=str, default="", help="The client ID to use when connecting to the broker. By default, a random client ID is generated by the broker."
    )
    parser.add_argument(
        "--clean_start", action="store_true", help="Start a clean session",
    )
    parser.add_argument(
        "--connect_timeout", type=int, default=10, help="Connection timeout in seconds",
    )
    parser.add_argument(
        "--username", type=str, default=None, help="The username for broker authentication"
    )
    parser.add_argument(
        "--password", type=str, default=None, help="The password for broker authentication (will be encoded as UTF-8)"
    )
    parser.add_argument(
        "--session_expiry_interval", type=int, help="The session expiry interval in seconds"
    )
    parser.add_argument(
        "--receive_maximum", type=int, help="Maximum number of QoS 1 and 2 messages that can be in-flight from the broker [1-65535]"
    )
    parser.add_argument(
        "--max_packet_size", type=int, help="Maximum packet size the client is willing to accept in bytes"
    )
    parser.add_argument(
        "--topic_alias_maximum", type=int, help="Maximum topic alias the client is willing to accept [0-65535]"
    )
    parser.add_argument(
        "--request_response_info", action="store_true", help="Request response information from the broker"
    )
    parser.add_argument(
        "--request_problem_info", action="store_true", help="Request problem information from the broker"
    )
    return parser


def get_client(args: argparse.Namespace) -> Client:
    """Create and return an MQTT client connected to the specified address."""
    address = args.address
    client_id = args.client_id
    clean_start = args.clean_start
    connect_timeout = args.connect_timeout
    username = args.username
    password = args.password
    props = MQTTConnectProps()
    prop_map = {
        "session_expiry_interval": "SessionExpiryInterval",
        "receive_maximum": "ReceiveMaximum",
        "max_packet_size": "MaximumPacketSize",
        "topic_alias_maximum": "TopicAliasMaximum",
        "request_response_info": "RequestResponseInformation",
        "request_problem_info": "RequestProblemInformation",
    }
    for arg_name, prop_name in prop_map.items():
        value = getattr(args, arg_name, None)
        if value is not None:
            setattr(props, prop_name, value)
    client = Client()
    try:
        client.connect(
            address,
            client_id=client_id,
            clean_start=clean_start,
            username=username,
            password=password.encode("utf-8") if password is not None else None,
            connect_properties=props,
        )
        if not client.loop_until_connected(timeout=connect_timeout):
            raise TimeoutError("Connection to broker timed out")
        return client
    except Exception as exc:
        raise BrokerConnectionError(f"Failed to connect to MQTT broker at {address}") from exc


class Command(metaclass=ABCMeta):
    """Abstract base class for CLI commands."""

    @classmethod
    @abstractmethod
    def register(cls, sub_parsers: SubParsersT) -> None:
        """Add command-specific arguments to the parser."""

    @classmethod
    @abstractmethod
    def execute(cls, args: argparse.Namespace) -> None:
        """Execute the command with the given arguments."""



"""Primitives for encoding and decoding MQTT packet fields."""

import struct
from typing import Final

from .error import MQTTError
from .mqtt_spec import MQTTReasonCode

# Maximum variable integer value.
MAX_VARINT: Final = 268435455


UInt8Packer: Final = struct.Struct("B")
UInt16Packer: Final = struct.Struct(">H")
UInt32Packer: Final = struct.Struct(">L")


def encode_bool(x: bool) -> bytes:
    """Encode a boolean to a buffer."""
    # MQTT booleans are encoded as a single byte with a value of 0 or 1.
    return b"\x01" if x else b"\x00"


def decode_bool(data: memoryview) -> tuple[bool, int]:
    """Decode a boolean from a buffer.

    Returns a tuple of the decoded boolean and the number of bytes consumed."""
    try:
        if data[0] not in (0, 1):
            raise MQTTError("Invalid boolean value", MQTTReasonCode.ProtocolError)
        return bool(data[0]), 1
    except MQTTError:
        raise
    except Exception as e:
        raise MQTTError("Failed to decode boolean from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_uint8(x: int) -> bytes:
    """Encode an 8-bit integer to a buffer."""
    return UInt8Packer.pack(x)


def decode_uint8(data: memoryview) -> tuple[int, int]:
    """Decode an 8-bit integer from a buffer.

    Returns a tuple of the decoded integer and the number of bytes consumed."""
    try:
        return UInt8Packer.unpack_from(data)[0], 1
    except Exception as e:
        raise MQTTError("Failed to decode byte from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_uint16(x: int) -> bytes:
    """Encode a 16-bit integer to a buffer."""
    return UInt16Packer.pack(x)


def decode_uint16(data: memoryview) -> tuple[int, int]:
    """Decode a 16-bit integer from a buffer.

    Returns a tuple of the decoded integer and the number of bytes consumed."""
    try:
        return UInt16Packer.unpack_from(data)[0], 2
    except Exception as e:
        raise MQTTError("Failed to decode 16-bit integer from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_uint32(x: int) -> bytes:
    """Encode a 32-bit integer to a buffer."""
    return UInt32Packer.pack(x)


def decode_uint32(data: memoryview) -> tuple[int, int]:
    """Decode a 32-bit integer from a buffer.

    Returns a tuple of the decoded integer and the number of bytes consumed."""
    try:
        return UInt32Packer.unpack_from(data)[0], 4
    except Exception as e:
        raise MQTTError("Failed to decode 32-bit integer from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_string(s: str) -> bytes:
    """Encode a UTF-8 string to a buffer."""
    data = s.encode("utf-8")
    return UInt16Packer.pack(len(data)) + data


def decode_string(data: memoryview) -> tuple[str, int]:
    """Decode a UTF-8 string from a buffer.

    Returns a tuple of the decoded string and the number of bytes consumed."""
    try:
        length = UInt16Packer.unpack_from(data)[0]
        if length > len(data) - 2:
            raise ValueError("String data underrun")
        # Strict UTF-8 decoding will catch any invalid UTF-8 sequences.
        # This is important for MQTT, as invalid sequences are not allowed.
        s = str(data[2:2 + length], encoding="utf-8", errors="strict")
        # The only other invalid character is the null character.
        if "\u0000" in s:
            raise ValueError("Unicode null character in string")
        return s, length + 2
    except Exception as e:
        raise MQTTError("Failed to decode string from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_string_pair(values: tuple[str, str]) -> bytes:
    """Encode a UTF-8 string pair to a buffer."""
    left = values[0].encode("utf-8")
    right = values[1].encode("utf-8")
    return UInt16Packer.pack(len(left)) + left + UInt16Packer.pack(len(right)) + right


def decode_string_pair(data: memoryview) -> tuple[tuple[str, str], int]:
    """Decode a UTF-8 string pair from a buffer.

    Returns a tuple of the decoded string pair and the number of bytes consumed."""
    try:
        key, key_length = decode_string(data)
        value, value_length = decode_string(data[key_length:])
        return (key, value), key_length + value_length
    except Exception as e:
        raise MQTTError("Failed to decode string pair from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_binary(data: bytes) -> bytes:
    """Encode binary data to a buffer."""
    return UInt16Packer.pack(len(data)) + data


def decode_binary(data: memoryview) -> tuple[bytes, int]:
    """Decode binary data from a buffer.

    Returns a tuple of the decoded data and the number of bytes consumed."""
    try:
        length = UInt16Packer.unpack_from(data)[0]
        if length > len(data) - 2:
            raise ValueError("Binary data underrun")
        return data[2:2 + length].tobytes("A"), length + 2
    except Exception as e:
        raise MQTTError("Failed to decode binary data from buffer", MQTTReasonCode.MalformedPacket) from e


def encode_varint(x: int) -> bytes:
    """Encode a variable length integer to a buffer."""
    # PEP20 told me that flat is better than nested, and I agree.
    if x < 0:
        raise ValueError(f"Value {x} must be greater than 0")
    if x <= 127:
        return x.to_bytes(1, "big")
    if x <= 16383:
        return bytes([
            (x & 0x7F) | 0x80,
            x >> 7
        ])
    if x <= 2097151:
        return bytes([
            (x & 0x7F) | 0x80,
            ((x >> 7) & 0x7F) | 0x80,
            x >> 14
        ])
    if x <= MAX_VARINT:
        return bytes([
            (x & 0x7F) | 0x80,
            ((x >> 7) & 0x7F) | 0x80,
            ((x >> 14) & 0x7F) | 0x80,
            x >> 21
        ])
    raise ValueError(f"Value {x} exceeds maximum {MAX_VARINT}")


def decode_varint(data: memoryview) -> tuple[int, int]:
    """Decode a variable length integer from a buffer.

    Returns a tuple of the decoded integer and the number of bytes consumed."""
    try:
        result = 0
        mult = 1
        sz = 0
        for byte in data:
            sz += 1
            result += byte % 0x80 * mult
            if byte < 0x80:
                return result, sz
            if sz >= 4:
                raise ValueError("Varint overflow")
            mult *= 0x80
        raise ValueError("Varint underrun")
    except Exception as e:
        raise MQTTError("Failed to decode varint from buffer", MQTTReasonCode.MalformedPacket) from e



from __future__ import annotations

import ssl
from typing import cast, Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .mqtt_handshake_connect import MQTTHandshakeConnectState
from .websocket_handshake_request import WebsocketHandshakeRequestState
from ..types import ConnectParams, StateData, StateEnvironment
from ...logger import get_logger

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.tls_handshake")


class TLSHandshakeState(FSMState):
    """Performing TLS handshake with the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        if params.tls_context is not None:
            tls_context = params.tls_context
        else:
            tls_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            tls_context.minimum_version = ssl.TLSVersion.TLSv1_3
            tls_context.maximum_version = ssl.TLSVersion.TLSv1_3
        state_data.sock = tls_context.wrap_socket(
            state_data.sock,
            server_hostname=params.tls_hostname if params.tls_hostname else params.address.host,
            do_handshake_on_connect=False,
        )
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("TLS handshake timeout")
            fsm.change_state(ClosedState)
            return True

        timeout = state_data.timeout.get_timeout(max_wait)
        sock = cast(ssl.SSLSocket, state_data.sock)
        try:
            logger.debug("trying TLS handshake")
            sock.do_handshake()
            if params.address.is_websocket():
                fsm.change_state(WebsocketHandshakeRequestState)
            else:
                fsm.change_state(MQTTHandshakeConnectState)
            return True
        except ssl.SSLWantReadError:
            logger.debug("TLS handshake wants read")
            with fsm.selector:
                fsm.selector.select(read=True, timeout=timeout)
        except ssl.SSLWantWriteError:
            logger.debug("TLS handshake wants write")
            with fsm.selector:
                fsm.selector.select(write=True, timeout=timeout)
        return False



import pytest

from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.packet import MQTTPublishPacket
from ohmqtt.topic_alias import AliasPolicy, MaxOutboundAliasError, InboundTopicAlias, OutboundTopicAlias


def make_packet(topic: str = "", alias: int | None = None) -> MQTTPublishPacket:
    """Create a MQTTPublishPacket with the given topic and alias."""
    packet = MQTTPublishPacket(topic=topic)
    if alias is not None:
        packet.properties.TopicAlias = alias
    return packet


def test_topic_alias_outbound() -> None:
    """Test the outbound topic alias lookup."""
    topic_alias = OutboundTopicAlias()
    topic_alias.max_alias = 10

    # Test with policy NEVER
    result = topic_alias.lookup("test/topic", AliasPolicy.NEVER)
    assert result == (0, False)

    # Test with policy TRY
    result = topic_alias.lookup("test/topic", AliasPolicy.TRY)
    assert result == (1, False)

    # Test with policy ALWAYS
    result = topic_alias.lookup("test/topic2", AliasPolicy.ALWAYS)
    assert result == (2, False)

    # Test with existing alias
    result = topic_alias.lookup("test/topic2", AliasPolicy.ALWAYS)
    assert result == (2, True)

    # Test with out of aliases
    topic_alias.max_alias = 2

    result = topic_alias.lookup("test/topic3", AliasPolicy.TRY)
    assert result == (0, False)

    with pytest.raises(MaxOutboundAliasError):
        topic_alias.lookup("test/topic4", AliasPolicy.ALWAYS)

    # Remove an alias
    topic_alias.pop()
    result = topic_alias.lookup("test/topic4", AliasPolicy.ALWAYS)
    assert result == (2, False)

    # Test reset
    topic_alias.reset()
    assert topic_alias.aliases == {}
    assert topic_alias.next_alias == 1
    assert topic_alias.max_alias == 2  # Should not reset


def test_topic_alias_inbound() -> None:
    """Test the inbound topic alias store."""
    topic_alias = InboundTopicAlias()

    # Unaliased, noop
    packet = make_packet("test/topic")
    topic_alias.handle(packet)
    assert packet == make_packet("test/topic")

    # No topic or alias, explode
    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet())
    assert exc.value.reason_code == MQTTReasonCode.ProtocolError

    # Test with max alias = 0
    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet(alias=1))
    assert exc.value.reason_code == MQTTReasonCode.TopicAliasInvalid

    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet("x", 1))
    assert exc.value.reason_code == MQTTReasonCode.TopicAliasInvalid

    # Test with max alias = 1
    topic_alias.max_alias = 1
    topic_alias.handle(make_packet("test/topic", 1))
    packet = make_packet(alias=1)
    topic_alias.handle(packet)
    assert packet.topic == "test/topic"

    # Trying to overwrite the same alias with a different topic
    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet("test/topic2", 1))
    assert exc.value.reason_code == MQTTReasonCode.TopicAliasInvalid

    # Trying to exceed the max alias
    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet("test/topic3", 2))
    assert exc.value.reason_code == MQTTReasonCode.TopicAliasInvalid

    # Lookup non-existing alias
    topic_alias.max_alias = 3
    with pytest.raises(MQTTError) as exc:
        topic_alias.handle(make_packet(alias=3))
    assert exc.value.reason_code == MQTTReasonCode.TopicAliasInvalid

    # Test reset
    topic_alias.reset()
    assert topic_alias.aliases == {}
    assert topic_alias.max_alias == 0



from __future__ import annotations

from typing import cast, Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .connected import ConnectedState
from ..decoder import ClosedSocketError
from ..types import ConnectParams, StateData, StateEnvironment
from ..wslib import OpCode, WebsocketError
from ...logger import get_logger
from ...mqtt_spec import MQTTPacketType
from ...packet import decode_packet, MQTTConnAckPacket, MQTTPacket

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.mqtt_handshake_connack")


class MQTTHandshakeConnAckState(FSMState):
    """Receiving MQTT CONNACK packet from the broker."""
    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNACK keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        want_read = False
        try:
            packet = cls.get_packet(state_data, params)
            want_read = packet is None
        except WebsocketError as exc:
            logger.error("WebSocket error while waiting for CONNACK: %s", exc)
            fsm.change_state(ClosedState)
            return True
        except ClosedSocketError:
            logger.exception("Socket was closed")
            fsm.change_state(ClosedState)
            return True

        if want_read:
            # Incomplete packet, wait for more data.
            with fsm.selector:
                timeout = state_data.timeout.get_timeout(max_wait)
                fsm.selector.select(read=True, timeout=timeout)
            return False

        if packet is not None and packet.packet_type == MQTTPacketType.CONNACK:
            packet = cast(MQTTConnAckPacket, packet)
            state_data.connack = packet
            if packet.properties.ServerKeepAlive is not None:
                state_data.keepalive.keepalive_interval = packet.properties.ServerKeepAlive
            fsm.change_state(ConnectedState)
            return True
        pt = packet.packet_type.name if packet is not None else "None"
        logger.error("Unexpected '%s' packet while waiting for CONNACK", pt)
        fsm.change_state(ClosedState)
        return True

    @classmethod
    def get_packet(cls, state_data: StateData, params: ConnectParams) -> MQTTPacket | None:
        """Get a packet from the socket.

        :returns: The received packet, or None if incomplete.
        :raises WebsocketError: A protocol error was encountered."""
        if params.address.is_websocket():
            decode_result = state_data.ws_decoder.decode(state_data.sock)
            if decode_result is None:
                return None
            opcode, payload = decode_result
            if opcode != OpCode.BINARY:
                raise WebsocketError(f"Unexpected WebSocket opcode {opcode.name} while waiting for CONNACK")
            return decode_packet(payload)
        return state_data.decoder.decode(state_data.sock)
